/**
 * A combobox to display notification app names and counts, as well as set the
 * application filter for the notification list.
 */
public class ApplicationComboBox : Gtk.ComboBox
{
  private Gtk.ListStore _list_store;

  private Gtk.CellRendererText _app_renderer;
  private CellRendererInteger  _count_renderer;

  /**
   * Columns of the Gtk.ListStore to hold the ComboBox values.
   */
  private enum Column {
    APP_NAME = 0,
    SEPARATOR,
    NOTIFICATIONS,
    COUNT
  }

  public ApplicationComboBox()
  {
    this._list_store = new Gtk.ListStore(Column.COUNT,
        typeof(string), // APP_NAME
        typeof(bool),   // SEPARATOR
        typeof(int)     // NOTIFICATIONS (count)
    );
    this.set_model(this._list_store);

    this._app_renderer = new Gtk.CellRendererText();
    this.pack_start(this._app_renderer, true);
    this.add_attribute(this._app_renderer, "text", Column.APP_NAME);

    this._count_renderer = new CellRendererInteger();
    this.pack_start(this._count_renderer, false);
    this.add_attribute(this._count_renderer, "integer", Column.NOTIFICATIONS);

    this.set_row_separator_func((model, iter) => {
      Value column_value;
      model.get_value(iter, Column.SEPARATOR, out column_value);
      return column_value.get_boolean();
    });

    this.clear_notifications();
  }

  /**
   * Returns the app_name for the given row.
   *
   * TODO: Same code as NotificationList, extract out various functions to pull
   * values from list stores into a separate namespace.
   *
   * @param iter the row to extract the app_name
   *
   * @return the app_name string for the given iter
   */
  private string get_app_name(Gtk.TreeIter iter)
  {
    Value column_value;
    this._list_store.get_value(iter, Column.APP_NAME, out column_value);

    return column_value.dup_string();
  }

  /**
   * Returns the active app name.
   *
   * @return the app_name string for the active iter
   */
  public string get_active_app_name()
  {
    Gtk.TreeIter iter;
    if(this.get_active_iter(out iter))
      return get_app_name(iter);
    else {
      warning("Invalid active row %d", this.get_active());
      return "";
    }
  }

  /**
   * Decrements the notification count for the given iter.
   *
   * @param iter the row to decrement
   */
  private void decrement_row(Gtk.TreeIter iter)
  {
    Value count;
    this._list_store.get_value(iter, Column.NOTIFICATIONS, out count);
    this._list_store.set_value(iter, Column.NOTIFICATIONS, count.get_int() - 1);
  }

  /**
   * Increments the notification count for the given iter.
   *
   * @param iter the row to increment
   */
  private void increment_row(Gtk.TreeIter iter)
  {
    Value count;
    this._list_store.get_value(iter, Column.NOTIFICATIONS, out count);
    this._list_store.set_value(iter, Column.NOTIFICATIONS, count.get_int() + 1);
  }

  /**
   * Updates the notification count for an app or adds another row to the
   * combobox.
   *
   * @param message the notification to add
   */
  public void add_notification(Notification message)
  {
    Gtk.TreeIter iter;

    /* Get the first row */
    if(!this._list_store.get_iter_first(out iter)) {
      warning("Application ComboBox should have at least one entry at all times, something is wrong here...");
      return;
    }

    /* Increment it */
    this.increment_row(iter);

    /* See if we already have the app_name */
    bool found_app = false;

    while(this._list_store.iter_next(ref iter)) {
      if(this.get_app_name(iter) == message.app_name) {
        found_app = true;
        /* We do, so just increment it */
        this.increment_row(iter);
        break;
      }
    }

    /* We don't, so add a new row */
    if(!found_app) {
      int length = this._list_store.iter_n_children(null);

      /* We don't have the separator yet, so add it */
      if(length == 1) {
        this._list_store.insert_with_values(null, length++,
          Column.APP_NAME, "",
          Column.SEPARATOR, true,
          Column.NOTIFICATIONS, 0);
      }

      this._list_store.insert_with_values(null, length,
        Column.APP_NAME, message.app_name,
        Column.SEPARATOR, false,
        Column.NOTIFICATIONS, 1);
    }
  }

  /**
   * Updates the notification count for an app to indicate a notification was
   * removed.
   *
   * @param message the notification to remove
   */
  public void remove_notification(Notification message)
  {
    Gtk.TreeIter iter;

    /* Get the first row */
    if(!this._list_store.get_iter_first(out iter)) {
      warning("Application ComboBox should have at least one entry at all times, something is wrong here...");
      return;
    }

    /* Decrement it */
    this.decrement_row(iter);

    /* Find the row and decrement it */
    while(this._list_store.iter_next(ref iter)) {
      if(this.get_app_name(iter) == message.app_name) {
        this.decrement_row(iter);
        return;
      }
    }

    /* We didn't find the row, something is wrong */
    warning("Notification missing: (%s) (%s) from (%s)", message.summary, message.body, message.app_name);
  }


  /**
   * Clears the combobox and adds the initial 'All Applications' row.
   */
  public void clear_notifications()
  {
    this._list_store.clear();
    this._list_store.insert_with_values(null, 0,
      Column.APP_NAME, _("All Applications"), 
      Column.SEPARATOR, false, 
      Column.NOTIFICATIONS, 0);
    this.set_active(0);
  }
}
