/**
 * Abstract base class for a generic notification filter framework. Will be
 * used to filter the notification list and also to discard notifications
 * before they arrive.
 */
public abstract class Filter : Object
{
  /**
   * Returns true if this filter matches the given notification.
   *
   * @param message the notification to match
   *
   * @return true if the notification matches
   */
  public abstract bool match(Notification message);
}

/**
 * A filter group applies a series of filters and returns the result. This
 * is the abstract base class.
 *
 * FIXME: It is a bit awkward to be calling methods on a null object, even
 * though it does make sense with the way GList works. Perhaps switch to libgee
 * to get real classes for various collections. 
 *
 * @see AndFilterGroup
 * @see OrFilterGroup
 */
public abstract class FilterGroup : Filter
{
  protected List<Filter> filters = null;

  /**
   * Adds a filter to the group.
   *
   * @param f the filter to add
   */
  public void add_filter(Filter f)
  {
    this.filters.append(f);
  }

  /**
   * Remove all filters from the group.
   */
  public void clear_filters()
  {
    this.filters = null;
  }
}

/**
 * This filter group returns true on match when all filters match a
 * notification. The empty group is considered true.
 */
public class AndFilterGroup : FilterGroup
{
  /**
   * Returns true when all filters in this group match the notification.
   *
   * @param message the notification to match
   *
   * @return true if all filters match
   */
  public override bool match(Notification message)
  {
    foreach(var filter in this.filters) {
      if(!filter.match(message))
        return false;
    }
    return true;
  }
}

/**
 * This filter group returns true on match when any filter matches a
 * notification. The empty group is considered false.
 */
public class OrFilterGroup : FilterGroup
{
  /**
   * Returns true when any filter in this group matches the notification.
   *
   * @param message the notification to match
   *
   * @return true if any filter matches
   */
  public override bool match(Notification message)
  {
    foreach(var filter in this.filters) {
      if(filter.match(message))
        return true;
    }
    return false;
  }
}

/**
 * Matches notifications with is_empty = true.
 */
public class EmptyFilter : Filter
{
  /**
   * Returns true when the notification is empty.
   *
   * @param message the notification to match
   *
   * @return true if empty
   */
  public override bool match(Notification message)
  {
    return message.is_empty();
  }
}

/**
 * Filters by exact match on a notification property.
 */
public class PropertyFilter : Filter
{
  /**
   * Using an enumeration of properties to avoid runtime checking.
   */
  public enum PropertyType
  {
    APP_NAME,
    APP_ICON,
    SUMMARY,
    BODY
  }

  protected PropertyType property;
  protected string       pattern;

  /**
   * Creates a new property filter.
   *
   * @param property the notification property to match
   * @param pattern the string that must exactly match the property
   */
  public PropertyFilter(PropertyType property, string pattern)
  {
    this.property = property;
    this.pattern = pattern;
  }

  /**
   * Extracts the property value from a notification and converts it to a
   * string.
   *
   * @param message the notification with the property
   *
   * @return the property as a string
   */
  protected string extract_property(Notification message)
  {
    switch(this.property) {
      case PropertyType.APP_NAME:
        return message.app_name;
      case PropertyType.APP_ICON:
        return message.app_icon;
      case PropertyType.SUMMARY:
        return message.summary;
      case PropertyType.BODY:
        return message.body;
      default:
        warning("Attempt to extract invalid property %d", this.property);
        return "";
    }
  }

  /**
   * Returns true when property matches pattern in this notification.
   *
   * @param message the notification to match
   *
   * @return true if an exact match
   */
  public override bool match(Notification message)
  {
    return extract_property(message) == this.pattern;
  }
}

/**
 * Filters with a hashed set of strings.
 *
 */
public class HashPropertyFilter : PropertyFilter
{
  /* Let key == value for optimized set version */
  protected HashTable<string, string> _table;

  public HashPropertyFilter(PropertyFilter.PropertyType property)
  {
    base(property, "");

    this._table = new HashTable<string, string>(str_hash, str_equal);
  }

  /**
   * Clears the set and inserts an array of strings.
   *
   * @param items the strings to add
   */
  public void set_values(string[] values)
  {
    /* Need to lock the table when resetting the values because the
     * notification monitor runs in another thread, and will be attempting to
     * match against the table.
     */
    lock(this._table) {
      /* Clear the table */
      this._table.remove_all();

      for(int i = 0; i < values.length; i++) {
        /* Values aren't stored if key == value */
        this._table.insert(values[i], values[i]);
      }
    }
  }
  
  /**
   * Returns true when the property is contained in the string set.
   * 
   * @param message the notification to match
   *
   * @return true if the property exists in the set
   */
  public override bool match(Notification message)
  {
    /* Lock the table in case the values are being updated when a notification
     * arrives.
     */
    lock(this._table) {
      string key = extract_property(message);
      return this._table.lookup_extended(key, null, null);
    }
  }
}

/**
 * Filters by regex on a notification property. Uses the syntax from:
 * http://developer.gnome.org/glib/2.29/glib-regex-syntax.html
 */
public class RegexPropertyFilter : PropertyFilter
{
  protected Regex regex;

  /**
   * Creates a new regex filter for the property.
   *
   * @param property the notification property to match
   * @param pattern the regex string
   */
  public RegexPropertyFilter(PropertyFilter.PropertyType property, string pattern)
  {
    base(property, pattern);

    try {
      this.regex = new Regex(pattern);
    }
    catch(RegexError e) {
      warning("Failed to parse regex pattern '%s': %s", pattern, e.message);
      this.regex = null;
    }
  }

  /**
   * Returns true when the property matches the regex pattern in this
   * notification.
   * 
   * @param message the notification to match
   *
   * @return true if the regex matches
   */
  public override bool match(Notification message)
  {
    if(this.regex == null)
      return false;
    else
      return this.regex.match(extract_property(message));
  }
}
