/**
 * Collection of functions to load icons.
 */
namespace Icon
{
  /**
   * Holds image data included in a notification after it is converted from a
   * Variant.
   */
  public class ImageData : Object
  {
    public int  width;
    public int  height;
    public int  rowstride;
    public bool has_alpha;
    public int  bits_per_sample;
    public int  channels;

    public uint8[] data;

    /**
     * Creates a new pixbuf from this image data and scales it to the given
     * size.
     *
     * @param size the size to scale the image data
     *
     * @return a new pixbuf or null if it fails
     */
    public Gdk.Pixbuf? to_pixbuf(int size)
    {
      /**
       * Pixbuf.from_data only supports 3 channels and 8 bits per sample, or 4
       * channels if has_alpha is true.
       */
      if((has_alpha && channels != 4) || (!has_alpha && channels != 3) || bits_per_sample != 8) {
        warning("Invalid image with %d channels and %d bits per sample with alpha %s", channels,
          bits_per_sample, (has_alpha) ? "true" : "false");
        return null;
      }

      Gdk.Pixbuf pixbuf = new Gdk.Pixbuf.from_data(data, Gdk.Colorspace.RGB,
        has_alpha, bits_per_sample, width, height, rowstride, null);

      return pixbuf.scale_simple(size, size, Gdk.InterpType.BILINEAR);
    }
  }

  /**
   * Loads an icon from either a file:// URI or from the default icon theme.
   * (The two methods supported in the Desktop Notifications Spec for app_icon)
   *
   * @param name the name or uri of the icon
   * @param size the size to scale the icon
   *
   * @return a pixbuf or null if no icon was found
   */
  public static Gdk.Pixbuf? load(string name, int size)
  {
    if(name.strip().length == 0)
      return null;
    else if(name.has_prefix("file://"))
      return Icon.load_from_uri(name, size);
    else
      return Icon.load_from_icon_name(name, size);
  }

  /**
   * Loads an icon from the default icon theme.
   *
   * @param name the name of the icon
   * @param size the size to scale the icon
   *
   * @return a pixbuf or null if no icon was found
   */
  public static Gdk.Pixbuf? load_from_icon_name(string name, int size)
  {
    Gtk.IconTheme theme = Gtk.IconTheme.get_default();

    try {
      return theme.load_icon(name, size, Gtk.IconLookupFlags.FORCE_SIZE);
    }
    catch(Error e) {
      warning("Failed to load icon '%s' at size %d: %s", name, size, e.message);
      return null;
    }
  }

  /**
   * Loads an icon from a file:// uri.
   *
   * @param uri the uri of the icon
   * @param size the size to scale the icon
   *
   * @return a pixbuf or null if no icon was found
   */
  public static Gdk.Pixbuf? load_from_uri(string uri, int size)
  {
    string path = "";

    try {
      path = Filename.from_uri(uri);
    }
    catch(ConvertError e) {
      warning("Failed to convert uri '%s' to path: %s", uri, e.message);
      return null;
    }

    try {
      return new Gdk.Pixbuf.from_file_at_size(path, size, size);
    }
    catch(Error e) {
      warning("Failed to load file '%s': %s", path, e.message);
      return null;
    }
  }
}
