/**
 * Enables logging to both a file and the console using standard GLib functions
 * (debug/warning/error etc).
 */
public class Logging : Object
{
  private string     _log_path;
  private FileStream _log_file;

  private bool _log_debug      = true;
  private bool _log_to_console = true;

  /**
   * Configures logging to a file in the user cache directory (generally
   * ~/.cache/).
   *
   * @param log_filename the filename for the log file
   */
  public Logging(string log_filename)
  {
    string cache_dir = Environment.get_user_cache_dir();
    this._log_path = Path.build_filename(cache_dir, log_filename, null);

    DirUtils.create_with_parents(cache_dir, 0700);
    this._log_file = FileStream.open(this._log_path, "w");

    Log.set_default_handler(this.logging_cb);

    if(this._log_file == null)
      warning("Failed to open log file: %s", this._log_path);
    else {
      debug("Logging to %s", this._log_path);
    }
  }

  /**
   * Logs messages to both a file and to the console.
   *
   * @param domain the domain of the message [currently ignored]
   * @param level the log level (debug, warning, error, ...)
   * @param message the log message
   */
  private void logging_cb(string? domain, LogLevelFlags level, string message)
  {
    string prefix = "LOG:";
    unowned FileStream console = stdout;

    switch(level) {
      case LogLevelFlags.LEVEL_ERROR:
        prefix = "ERROR:";
        console = stderr;
        break;
      case LogLevelFlags.LEVEL_CRITICAL:
        prefix = "CRITICAL:";
        console = stderr;
        break;
      case LogLevelFlags.LEVEL_WARNING:
        prefix = "WARNING:";
        console = stderr;
        break;
      case LogLevelFlags.LEVEL_MESSAGE:
        prefix = "MESSAGE:";
        console = stderr;
        break;
      case LogLevelFlags.LEVEL_INFO:
        prefix = "INFO:";
        break;
      case LogLevelFlags.LEVEL_DEBUG:
        /* Skip debug messages when debugging is disabled */
        if(!this._log_debug) return;

        prefix = "DEBUG:";
        break;
    }

    if(this._log_file != null) {
      this._log_file.printf("%s %s\n", prefix, message);
      this._log_file.flush();
    }

    if(this._log_to_console) {
      console.printf("%s %s\n", prefix, message);
    }
  }

  /**
   * Property to determine if debug messages are logged.
   */
  public bool log_debug
  {
    get { return this._log_debug; }
    set {
      this._log_debug = value;
    }
  }

  /**
   * Property to determine if messages are logged to the console.
   */
  public bool log_to_console
  {
    get { return this._log_to_console; }
    set {
      this._log_to_console = value;
    }
  }
}
