/**
 * Context menu for the notification list.
 */
public class NotificationListMenu : Gtk.Menu
{
  /**
   * Creates a new context menu populated with the items from the given
   * notification.
   *
   * @param message the notification for this context menu
   */
  public NotificationListMenu(Notification message)
  {
    /* Copy Menu */
    Gtk.MenuItem copy_item = new Gtk.MenuItem.with_label(_("Copy"));
    copy_item.show();
    this.append(copy_item);

    Gtk.Menu copy_menu = new Gtk.Menu();
    copy_menu.append(create_copy_item(_("Summary"), message.summary));
    copy_menu.append(create_copy_item(_("Body"), message.body));

    /* Only show Icon menu item if the notification has an icon */
    Gdk.Pixbuf pixbuf = message.load_icon(NotificationList.DEFAULT_ICON_SIZE);
    if(pixbuf != null)
      copy_menu.append(create_copy_image_item(_("Icon"), pixbuf));

    copy_menu.append(create_copy_item(_("Application"), message.app_name));

    /* Only show Embedded Links menu if the notification has links */
    string[] links = message.get_embedded_links();
    if(links.length > 0) {
      Gtk.MenuItem copy_links_item = new Gtk.MenuItem.with_label(_("Embedded Links"));
      copy_links_item.show();
      copy_links_item.submenu = create_copy_links_menu(links);
      copy_menu.append(copy_links_item);
    }

    copy_item.submenu = copy_menu;

    /* Visit Menu (hidden if empty) */
    if(links.length > 0) {
      Gtk.MenuItem visit_item = new Gtk.MenuItem.with_label(_("Visit"));
      visit_item.show();
      visit_item.submenu = create_visit_links_menu(links);
      this.append(visit_item);

    }

    /* Make sure to destroy the menu when we're done */
    this.selection_done.connect((menu) => {
      menu.destroy();
    });
  }

  /**
   * Creates an item for the copy menu with the given label.
   *
   * @param label the label of this copy menuitem
   * @param text the text that is copied to the clipboard on activate
   *
   * @return a new menuitem
   */
  private Gtk.MenuItem create_copy_item(string label, string text)
  {
    Gtk.MenuItem item = new Gtk.MenuItem.with_label(label);
    item.activate.connect((menuitem) => {
      debug("Copy '%s' to clipboard", text);
      unowned Gtk.Clipboard clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD);
      clipboard.set_text(text, -1);
    });
    item.show();
    return item;
  }

  /**
   * Creates an item for the copy menu with the given image.
   *
   * @param label the label of this copy menuitem
   * @param pixbuf the pixbuf to copy to the clipboard
   *
   * @return a new menuitem
   */
  private Gtk.MenuItem create_copy_image_item(string label, Gdk.Pixbuf pixbuf)
  {
    Gtk.MenuItem item = new Gtk.MenuItem.with_label(label);
    item.activate.connect((menuitem) => {
      debug("Copy image to clipboard");
      unowned Gtk.Clipboard clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD);
      clipboard.set_image(pixbuf);
    });
    item.show();
    return item;
  }

  /**
   * Creates a menu to copy link urls to the clipboard.
   * 
   * @param links the web link urls from the notification
   *
   * @return a new menu
   */
  private Gtk.Menu create_copy_links_menu(string[] links)
  {
    Gtk.Menu menu = new Gtk.Menu();

    foreach(var link in links)
      menu.append(create_copy_link_item(link));

    return menu;
  }

  /**
   * Creates a menu to visit link urls.
   *
   * FIXME: Merge with create_copy_links_menu
   *
   * @param links the web link urls to visit
   *
   * @return a new menu
   */
  private Gtk.Menu create_visit_links_menu(string[] links)
  {
    Gtk.Menu menu = new Gtk.Menu();

    foreach(var link in links)
      menu.append(create_visit_link_item(link));

    return menu;
  }

  /**
   * Creates a new possibly truncated menuitem. Will truncate if greater than
   * 32 characters and set the tooltip to the full string.
   *
   * @param label the menuitem's label
   *
   * @return a new menuitem
   */
  private Gtk.MenuItem create_truncated_item(string label)
  {
    string new_label = label;
    bool truncated = false;
    if(new_label.length > 32) {
      new_label = new_label[0:32] + "...";
      truncated = true;
    }
    Gtk.MenuItem item = new Gtk.MenuItem.with_label(new_label);
    if(truncated) {
      item.set_has_tooltip(true);
      item.set_tooltip_text(label);
    }
    item.show();
    return item;
  }

  /**
   * Creates a new copy link menu item.
   *
   * @param link the link to copy
   *
   * @return a new menuitem
   */
  private Gtk.MenuItem create_copy_link_item(string link)
  {
    Gtk.MenuItem item = create_truncated_item(link);
    item.activate.connect((menuitem) => {
      debug("Copy '%s' to clipboard", link);
      unowned Gtk.Clipboard clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD);
      clipboard.set_text(link, -1);
    });
    return item;
  }

  /**
   * Creates a new visit link menu item. Uses Gtk.show_uri.
   *
   * @param link the link to visit
   *
   * @return a new menuitem
   */
  private Gtk.MenuItem create_visit_link_item(string link)
  {
    Gtk.MenuItem item = create_truncated_item(link);
    item.activate.connect((menuitem) => {
      debug("Visit '%s'", link);
      try {
        Gtk.show_uri(null, link, Gdk.CURRENT_TIME);
      }
      catch(Error e) {
        /* FIXME: Show error dialog? */
        warning("Failed to visit '%s'", link);
      }
    });
    return item;
  }
}
