/**
 * A Gtk.TreeView to display notifications.
 */
public class NotificationList : Gtk.TreeView
{
  private Gtk.ListStore _list_store;

  private Gtk.TreeModelFilter _tree_model;

  private Gtk.TreeViewColumn     _column;
  private Gtk.CellRendererPixbuf _icon_renderer;
  private Gtk.CellRendererText   _text_renderer;

  private FilterGroup _filters;

  /**
   * Columns of the Gtk.ListStore to hold the TreeView values.
   */
  private enum Column {
    APP_NAME = 0,
    PIXBUF,
    SUMMARY,
    BODY,
    TIMESTAMP,
    NOTIFICATION,
    COUNT
  }

  /**
   * Template for the markup of a previously viewed message.
   */
  private const string read_template = "<b>%s</b>\n%s\n<small><i>%s %s <b>%s</b></i></small>";

  public const int DEFAULT_ICON_SIZE = 48;

  public NotificationList()
  {
    this._list_store = new Gtk.ListStore(Column.COUNT, 
        typeof(string),      // APP_NAME
        typeof(Gdk.Pixbuf),  // PIXBUF
        typeof(string),      // SUMMARY
        typeof(string),      // BODY
        typeof(DateTime),    // TIMESTAMP
        typeof(Notification) // NOTIFICATION
    );

    this._filters = new AndFilterGroup();

    this._tree_model = new Gtk.TreeModelFilter(this._list_store, null);
    this._tree_model.set_visible_func((model, iter) => {
      Notification message = get_notification(model, iter);
      if(message == null)
        return false;
      else
        return this._filters.match(message);
    });

    this.set_model(this._tree_model);
    this.set_headers_visible(false);

    this._column = new Gtk.TreeViewColumn();

    this._icon_renderer = new Gtk.CellRendererPixbuf();
    this._icon_renderer.xpad = 5;
    this._icon_renderer.ypad = 5;
    this._icon_renderer.xalign = 0;
    this._icon_renderer.yalign = 0;
    this._column.pack_start(this._icon_renderer, false);
    this._column.add_attribute(this._icon_renderer, "pixbuf", Column.PIXBUF);

    this._text_renderer = new Gtk.CellRendererText();
    this._text_renderer.xpad = 5;
    this._text_renderer.ypad = 5;
    this._text_renderer.wrap_mode = Pango.WrapMode.WORD_CHAR;
    this._column.pack_start(this._text_renderer, true);
    this._column.set_cell_data_func(this._text_renderer, this.update_markup_cb);

    this.append_column(this._column);

    /* Show context menu on button press */
    this.button_press_event.connect((widget, event) => {
      if(event.button == 3)
      {
        NotificationList list = widget as NotificationList;
        Gtk.TreePath path;
        if(list.get_path_at_pos((int)event.x, (int)event.y, out path, null, null, null))
        {
          list.set_cursor(path, null, false);

          Gtk.TreeModel model = list.get_model();
          Gtk.TreeIter iter;
          if(model.get_iter(out iter, path)) {
            Notification message = get_notification(model, iter);
            NotificationListMenu menu = new NotificationListMenu(message);
            menu.popup(null, null, null, event.button, event.time);
          }
        }
        return true;
      }
      return false;
    });
  }

  /**
   * Prepends a notification to the list.
   *
   * TODO: Merge duplicate messages.
   *
   * @param message the notification to prepend
   */
  public void add_notification(Notification message)
  {
    this._list_store.insert_with_values(null, 0,
      Column.APP_NAME, Markup.escape_text(message.app_name),
      Column.PIXBUF, message.load_icon(DEFAULT_ICON_SIZE),
      Column.SUMMARY, Markup.escape_text(message.summary),
      Column.BODY, Markup.escape_text(message.body),
      Column.TIMESTAMP, message.timestamp,
      Column.NOTIFICATION, message);
  }

  /**
   * Removes all notifications from the list.
   */
  public void clear_notifications()
  {
    this._list_store.clear();
  }

  /**
   * Removes the last notification in this list.
   *
   * @return the notification removed or null if the list is empty
   */
  public Notification? remove_last_notification()
  {
    int n = this.length - 1;
    Gtk.TreeIter iter;

    if(n < 0) return null;

    if(this._list_store.iter_nth_child(out iter, null, n)) {
      Notification message = get_notification(this._list_store, iter);
      this._list_store.remove(iter);
      return message;
    }

    return null;
  }

  /**
   * Adds a notification visibility filter.
   *
   * @param f the filter to add
   */
  public void add_filter(Filter f)
  {
    this._filters.add_filter(f);
    this._tree_model.refilter();
  }

  /**
   * Clears all notification visibility filters.
   */
  public void clear_filters()
  {
    this._filters.clear_filters();
    this._tree_model.refilter();
  }

  /**
   * Emits a row-changed signal for each row in the Gtk.ListStore.
   */
  public void all_rows_changed()
  {
    this._list_store.foreach((model, path, iter) => {
      this._list_store.row_changed(path, iter);
      return false;
    });
  }

  /**
   * Sets the wrap width for the Gtk.CellRendererText that contains the main
   * text of a notification. The width given should generally be the width of
   * the NotificationList's parent container.
   *
   * @param width the width allotted to each notification row
   */
  public void set_wrap_width(int width)
  {
    int icon_padding = (int)this._icon_renderer.xpad;
    int text_padding = (int)this._text_renderer.xpad;
    int wrap_width = width - 2*text_padding - 2*icon_padding - DEFAULT_ICON_SIZE;
    this._text_renderer.wrap_width = wrap_width;
    this.all_rows_changed();
  }

  /**
   * Extracts a string from a column in the Gtk.ListStore.
   *
   * @param model the model to extract
   * @param iter the row from which to extract the value
   * @param column the column in the row where the value lives
   *
   * @return a new string for the value of the row[column]
   */
  private string get_column_as_string(Gtk.TreeModel model, Gtk.TreeIter iter, Column column)
  {
    Value column_value;
    model.get_value(iter, column, out column_value);

    return column_value.dup_string();
  }

  /**
   * Extracts the notification from a column in the Gtk.ListStore.
   *
   * @param model the model to extract
   * @param iter the row from which to extract the notification
   *
   * @return the notification object
   */
  private Notification get_notification(Gtk.TreeModel model, Gtk.TreeIter iter)
  {
    Value column_value;
    model.get_value(iter, Column.NOTIFICATION, out column_value);

    return (Notification)column_value.get_object();
  }

  /**
   * Extracts the timestamp from a row and converts it to a string representing
   * the difference in time between when the notification arrived and when this
   * method was called. If the time period is greater than a day, a string with
   * the date of the notification timestamp in the locale preferred format is
   * returned.
   *
   * Examples:
   * less than a minute ago
   * 2 hours ago
   *
   * @param model the model to extract
   * @param iter the notification row
   *
   * @return a new string for the time difference
   */
  private string get_time_ago_in_words(Gtk.TreeModel model, Gtk.TreeIter iter)
  {
    Value column_value;
    model.get_value(iter, Column.TIMESTAMP, out column_value);

    DateTime previous = (DateTime)column_value.get_boxed();

    DateTime current = new DateTime.now_local();
    int64 seconds = current.to_unix() - previous.to_unix();

    if(seconds < 60)
      return _("less than a minute ago");
    else {
      int64 minutes = seconds / 60;
      if(minutes < 60)
        return ngettext("1 minute ago", "%lld minutes ago", (ulong)minutes).printf(minutes);
      else {
        int64 hours = minutes / 60;
        if(hours < 24)
          return ngettext("1 hour ago", "%lld hours ago", (ulong)hours).printf(hours);
      }
    }

    return previous.format("%c");
  }

  /**
   * This method is called whenever a cell needs to be redrawn. This allows
   * the text from get_time_ago_in_words to be updated at least every time
   * the notification window becomes visible, and also when the user mouses
   * over it.
   *
   * @param layout the Gtk.TreeViewColumn being updated [not used]
   * @param cell the Gtk.CellRendererText to update
   * @param model the Gtk.ListStore containing the rows
   * @param iter the row being updated
   *
   * @see get_time_ago_in_words
   */
  private void update_markup_cb(Gtk.CellLayout   layout, 
                                Gtk.CellRenderer cell, 
                                Gtk.TreeModel    model,
                                Gtk.TreeIter     iter)
  {
    string summary = get_column_as_string(model, iter, Column.SUMMARY);
    string body = get_column_as_string(model, iter, Column.BODY);
    string timestamp = Markup.escape_text(get_time_ago_in_words(model, iter));
    string app_name = get_column_as_string(model, iter, Column.APP_NAME);

    string message = read_template.printf(summary, 
      body, 
      timestamp, 
      _("from"),
      app_name);

    cell.set_property("markup", message);
  }

  /**
   * Returns the number of notifications in this list.
   */
  public int length
  {
    get {
      return this._list_store.iter_n_children(null);
    }
  }
}
