/**
 * Watches DBUS for org.freedesktop.Notifications.Notify messages and emits a
 * received signal with a Notification object when they arrive.
 *
 * @see Notification
 */
public class NotificationMonitor : Object
{
  private DBusConnection _connection = null;

  private FilterGroup        _filters;
  private HashPropertyFilter _blacklist_filter;

  private string[] _blacklist;

  /**
   * Emitted when a notification has been received.
   * 
   * @param message a Notification object
   */
  public signal void received(Notification message);

  public NotificationMonitor()
  {
    Bus.get.begin(BusType.SESSION, null, (obj, res) => {
      try {
        this._connection = Bus.get.end(res);
      }
      catch(IOError e) {
        error("Failed to connect to session bus: %s", e.message);
      }

      this.add_filter();
    });

    /* Filters to discard notifications */
    this._filters = new OrFilterGroup();
    this._filters.add_filter(new EmptyFilter());

    this._blacklist_filter = new HashPropertyFilter(PropertyFilter.PropertyType.APP_NAME);
    this._filters.add_filter(this._blacklist_filter);
  }

  ~NotificationMonitor()
  {
    if(this._connection != null)
      this._connection.close();
  }

  /**
   * Adds a DBUS message match string for org.freedesktop.Notifications.Notify
   * messsages and a message filter to process them when they arrive.
   */
  private void add_filter()
  {
    DBusMessage message = new DBusMessage.method_call("org.freedesktop.DBus",
        "/org/freedesktop/DBus", "org.freedesktop.DBus", "AddMatch");

    Variant body = new Variant.parsed("(%s,)",
        "eavesdrop=true,type='method_call',interface='org.freedesktop.Notifications',member='Notify'");

    message.set_body(body);

    try {
      this._connection.send_message(message, DBusSendMessageFlags.NONE, null);
    }
    catch(Error e) {
      error("Failed to add match string: %s", e.message);
    }

    this._connection.add_filter((connection, message, incoming) => {
      if(incoming) {
        if((message.get_message_type() == DBusMessageType.METHOD_CALL) &&
           (message.get_interface() == "org.freedesktop.Notifications") &&
           (message.get_member() == "Notify")) {
          Notification data = new Notification(message);
          /* Don't emit notifications that match the filters */
          if(this._filters.match(data))
            debug("Discarding notification from '%s'", data.app_name);
          else {
            Idle.add(() => { 
              this.received(data); 
              return false;
            });
          }
          message.unref();
          message = null;
        }
      }

      return message;
    });
  }

  /**
   * Property for the array of app_name strings to discard.
   */
  public string[] blacklist
  {
    get { return this._blacklist; }
    set {
      this._blacklist = value;
      this._blacklist_filter.set_values(value);
    }
  }
}
