/**
 * The dock window popped up when a user clicks the applet icon.
 */
public class NotificationWindow : Gtk.Window
{
  private Gtk.Box            _box;
  private Gtk.ScrolledWindow _scroll;
  
  private ApplicationComboBox _combobox;
  private NotificationList    _list;

  private int _notification_limit = 0;

  /* FIXME: Determine window size based on screen dimensions? */
  private const int WINDOW_SIZE = 400;


  public NotificationWindow()
  {
    this.set_type_hint(Gdk.WindowTypeHint.DOCK);
    this.set_decorated(false);
    this.set_resizable(false);
    this.set_size_request(WINDOW_SIZE, WINDOW_SIZE);

    this._box = new Gtk.Box(Gtk.Orientation.VERTICAL, 0);
    this._box.show();
    this.add(this._box);

    /* Scrolled window for the notification list */

    this._scroll = new Gtk.ScrolledWindow(null, null);
    this._scroll.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);
    this._scroll.show();
    this._box.pack_start(this._scroll, true, true, 0);

    Gtk.Adjustment adjust = this._scroll.get_vadjustment();
    adjust.changed.connect((adjust) => {
      adjust.set_value(0);
    });

    /* Combo box for the application filter */
    this._combobox = new ApplicationComboBox();
    this._combobox.show();
    this._box.pack_start(this._combobox, false, false, 0);

    /* TreeView to display the notifications */

    this._list = new NotificationList();
    this._list.set_wrap_width(WINDOW_SIZE);
    this._list.show();
    this._scroll.add(this._list);

    /* Update filters on changes to the combobox */
    this._combobox.changed.connect((combobox) => {
      /* Clear the old filters */
      debug("Clearing Application filters");
      this._list.clear_filters();
      if(combobox.get_active() > 0) {
        ApplicationComboBox app_combobox = combobox as ApplicationComboBox;
        string app_name = app_combobox.get_active_app_name();
        debug("Adding application filter for '%s'", app_name);
        PropertyFilter f = new PropertyFilter(PropertyFilter.PropertyType.APP_NAME, 
          app_name);
        this._list.add_filter(f);
      }
    });
  }

  /**
   * Adds a notification to the NotificationList.
   *
   * @param message the notification to add
   */
  public void add_notification(Notification message)
  {
    this._combobox.add_notification(message);
    this._list.add_notification(message);
    this.discard_old_notifications();
  }

  /**
   * Removes all notifications from this window's list.
   */
  public void clear_notifications()
  {
    this._combobox.clear_notifications();
    this._list.clear_notifications();
  }

  /**
   * Discards old notification if over the notification limit.
   */
  public void discard_old_notifications()
  {
    int limit = this.notification_limit;

    /* Unlimited */
    if(limit == 0) return;

    /* Remove old notifications until we're back to the limit */
    while(this._list.length > limit) {
      Notification message = this._list.remove_last_notification();
      this._combobox.remove_notification(message);
    }
  }

  /** 
   * Positions the window based on the position and orientation of the applet.
   * Adapted from position_calendar_popup in the GNOME clock applet.
   *
   * @param parent_x the x position of the applet
   * @param parent_y the y position of the applet
   * @param parent_width the width of the applet
   * @param parent_height the height of the applet
   * @param screen the screen to show the window on
   * @param orient the orientation of the panel containing the applet
   */
  public void align_with_parent(int parent_x, int parent_y, 
                                int parent_width, int parent_height,
                                Gdk.Screen screen, Panel.AppletOrient orient)
  {
    int x = parent_x;
    int y = parent_y;

    int width, height;
    this.get_size(out width, out height);

    Gdk.Rectangle monitor = {0, 0, screen.get_width(), screen.get_height()};

    /* Figure out which monitor we are on */
    for(int i = 0; i < screen.get_n_monitors(); i++) {
      Gdk.Rectangle rect;
      screen.get_monitor_geometry(i, out rect);

      if(x >= rect.x && x <= (rect.x + rect.width) &&
         y >= rect.y && y <= (rect.y + rect.height)) {
        monitor.x = rect.x;
        monitor.y = rect.y;
        monitor.width = rect.width;
        monitor.height = rect.height;
        break;
      }
    }

    /* Figure out the (x, y) coordinates of the window */
    switch(orient) {
      case Panel.AppletOrient.RIGHT: /* Panel on left */
        x += parent_width;
        if((y + height) > (monitor.y + monitor.height))
          y -= (y + height) - (monitor.y + monitor.height);
        break;
      case Panel.AppletOrient.LEFT:  /* Panel on right */
        x -= width;
        if((y + height) > (monitor.y + monitor.height))
          y -= (y + height) - (monitor.y + monitor.height);
        break;
      case Panel.AppletOrient.DOWN:  /* Panel on top */
        y += parent_height;
        if((x + width) > (monitor.x + monitor.width))
          x -= (x + width) - (monitor.x + monitor.width);
        break;
      case Panel.AppletOrient.UP:    /* Panel on bottom */
        y -= height;
        if((x + width) > (monitor.x + monitor.width))
          x -= (x + width) - (monitor.x + monitor.width);
        break;
      default:
        error("Invalid orient in align_with_parent: %d", orient);
    }

    this.set_gravity(Gdk.Gravity.NORTH_WEST);
    this.move(x, y);
  }

  /**
   * Property to determine the notification limit of this container.
   */
  public int notification_limit
  {
    get { return this._notification_limit; }
    set {
      debug("Updating notification limit to %d", value);
      this._notification_limit = value;
      this.discard_old_notifications();
    }
  }
}
