/**
 * An object representing a desktop notification.
 */
public class Notification : Object
{
  private string _app_name = "";
  private string _app_icon = "";
  private string _summary  = "";
  private string _body     = "";

  private DateTime _timestamp;

  /* Hints */
  private string _image_path = "";

  private Icon.ImageData _image_data = null;
  private Icon.ImageData _icon_data = null;

  private List<string> _embedded_links = null;

  /**
   * The DBusMessage layout for org.freedesktop.Notifications.Notify messages.
   *
   * Specification:
   * http://people.gnome.org/~mccann/docs/notification-spec/notification-spec-1.2.html
   */
  private enum Column {
    APP_NAME = 0,
    REPLACES_ID,
    APP_ICON,
    SUMMARY,
    BODY,
    ACTIONS,
    HINTS,
    EXPIRE_TIMEOUT,
    COUNT
  }

  private enum ImageColumn {
    WIDTH = 0,
    HEIGHT,
    ROWSTRIDE,
    HAS_ALPHA,
    BITS_PER_SAMPLE,
    CHANNELS,
    DATA,
    COUNT
  }

  /* Variant Helper Functions */

  /**
   * Gets a string from a Variant tuple at a column.
   *
   * @param tuple the Variant container
   * @param column the column to extract
   *
   * @return a new string for the value of the column
   */
  private static string get_string(Variant tuple, int column)
  {
    Variant child = tuple.get_child_value(column);
    if(!child.is_of_type(VariantType.STRING)) {
      warning("Invalid string column[%d] of type: %s", column, child.get_type_string());
      return "";
    }
    return child.dup_string();
  }

  /**
   * Gets an integer from a Variant tuple at a column.
   *
   * @param tuple the Variant container
   * @param column the column to extract
   *
   * @return an integer for the value of the column
   */
  private static int get_integer(Variant tuple, int column)
  {
    Variant child = tuple.get_child_value(column);
    if(!child.is_of_type(VariantType.INT32)) {
      warning("Invalid integer column[%d] of type: %s", column, child.get_type_string());
      return 0;
    }
    return child.get_int32();
  }

  /** Gets a boolean from a Variant type at a column.
   *
   * @param tuple the Variant container
   * @param column the column to extract
   *
   * @return a boolean for the value of the column
   */
  private static bool get_boolean(Variant tuple, int column)
  {
    Variant child = tuple.get_child_value(column);
    if(!child.is_of_type(VariantType.BOOLEAN)) {
      warning("Invalid boolean column[%d] of type: %s", column, child.get_type_string());
      return false;
    }
    return child.get_boolean();
  }

  /**
   * Gets a byte (uint8) from a Variant type at a column.
   *
   * @param tuple the Variant container
   * @param column the column to extract
   *
   * @return a byte for the value of the column
   */
  private static uint8 get_byte(Variant tuple, int column)
  {
    Variant child = tuple.get_child_value(column);
    if(!child.is_of_type(VariantType.BYTE)) {
      warning("Invalid byte column[%d] of type: %s", column, child.get_type_string());
      return 0;
    }
    return child.get_byte();
  }

  /**
   * Looks up a string in a Variant dictionary with a key.
   * 
   * @param dictionary the Variant dictionary
   * @param key the key
   *
   * @return a new string for the key or the empty string
   */
  private static string lookup_string(Variant dictionary, string key)
  {
    Variant child = dictionary.lookup_value(key, null);

    if(child == null)
      return "";

    if(!child.is_of_type(VariantType.STRING)) {
      warning("Invalid string value for key '%s': %s", key, child.get_type_string());
      return "";
    }

    return child.dup_string();
  }

  /**
   * Looks up image data in a Variant dictionary. Currently supported keys are
   * image-data and the deprecated icon_data (but this function doesn't check).
   *
   * @param dictionary the Variant dictionary
   * @param key the image data key
   *
   * @return the image data or null if there are errors
   * 
   * @see Icon.ImageData
   */
  public static Icon.ImageData? lookup_image_data(Variant dictionary, string key)
  {
    Variant child = dictionary.lookup_value(key, null);

    if(child == null)
      return null;

    if(!child.is_of_type(VariantType.TUPLE)) {
      warning("Invalid image data '%s' of type: %s", key, child.get_type_string());
      return null;
    }

    size_t length = child.n_children();
    if(length != ImageColumn.COUNT) {
      // FIXME: Log more information about the invalid data
      warning("Invalid image data '%s' of length: %d", key, (int)length);
      return null;
    }

    Icon.ImageData data = new Icon.ImageData();

    data.width = get_integer(child, ImageColumn.WIDTH);
    data.height = get_integer(child, ImageColumn.HEIGHT);
    data.rowstride = get_integer(child, ImageColumn.ROWSTRIDE);
    data.has_alpha = get_boolean(child, ImageColumn.HAS_ALPHA);
    data.bits_per_sample = get_integer(child, ImageColumn.BITS_PER_SAMPLE);
    data.channels = get_integer(child, ImageColumn.CHANNELS);

    /* Get the image data */
    Variant column_value = child.get_child_value(ImageColumn.DATA);

    if(!column_value.is_of_type(VariantType.ARRAY)) {
      warning("Invalid image data '%s' (data column) of type: %s", key, column_value.get_type_string());
      return null;
    }

    /* FIXME: This is likely going to be incredibly slow, however I am not sure
     * of any other way to safely get the data from the Variant. The Desktop
     * Notifications Specification lists the format as 'ay', which is what we
     * seem to be getting. The problem is that this is a BYTESTRING and
     * according to the documentation for g_variant_get_bytestring, the string
     * is expected to be null-terminated. Since the image probably has lots of
     * 0 bytes (and my test image starts with many of them), I can't get the
     * full array of data with this method.
     */
    data.data = new uint8[column_value.n_children()];
    int i;
    for(i = 0; i < column_value.n_children(); i++)
      data.data[i] = get_byte(column_value, i);

    return data;
  }

  /**
   * Constructs a new Notification from a DBusMessage.
   *
   * @param message the DBusMessage to parse
   */
  public Notification(DBusMessage message)
  {
    Variant body = message.get_body();

    /* Log raw message */
    debug("Creating new notification:\n%s", message.print(1));

    if(!body.is_of_type(VariantType.TUPLE)) {
      warning("Invalid message of type: %s", body.get_type_string());
      return;
    }

    size_t length = body.n_children();
    if(length != Column.COUNT) {
      // FIXME: Log more information about the invalid message
      warning("Invalid message of length: %d", (int)length);
      return;
    }

    this._app_name = get_string(body, Column.APP_NAME);
    this._app_icon = get_string(body, Column.APP_ICON);
    this._summary  = get_string(body, Column.SUMMARY);
    this._body     = get_string(body, Column.BODY);

    this._timestamp = new DateTime.now_local();

    /* Hints */
    Variant hints = body.get_child_value(Column.HINTS);

    if(!hints.is_of_type(VariantType.DICTIONARY)) {
      warning("Invalid hints of type: %s", hints.get_type_string());
      return;
    }

    this._image_path = lookup_string(hints, "image-path");
    this._image_data = lookup_image_data(hints, "image_data");
    this._icon_data = lookup_image_data(hints, "icon_data");

    extract_embedded_links(); 
  }

  /**
   * Create a notification from data instead of from a DBusMessage. This is for
   * testing support.
   *
   * @param app_name the notification app_name
   * @param app_icon the notification app_icon
   * @param summary  the notification summary
   * @param body     the notification body
   */
  public Notification.from_data(string app_name, 
                                string app_icon, 
                                string summary,
                                string body)
  {
    this._app_name = app_name;
    this._app_icon = app_icon;
    this._summary = summary;
    this._body = body;

    this._timestamp = new DateTime.now_local();

    extract_embedded_links(); 
  }

  /**
   * Extracts embedded links from the body of the notification.
   */
  private void extract_embedded_links()
  { 
    /** 
     * Extract embedded links from the notification body. Links are only
     * considered if they begin with http:// or www., and are not validated.
     */
    string[] tokens = this._body.split_set(" \t\r\n");
    foreach(var token in tokens) {
      if(token.has_prefix("http://")) {
        this._embedded_links.append(token);
      }
      else if(token.has_prefix("www.")) {
        this._embedded_links.append("http://" + token);
      }
    }
  }

  /**
   * Returns an array of urls.
   *
   * @return a newly created array of links
   */
  public string[] get_embedded_links()
  {
    string[] result = new string[this._embedded_links.length()];
    int i = 0;
    foreach(var link in this._embedded_links)
      result[i++] = link;
    return result;
  }

  /**
   * Loads the icon using the priorities from the Desktop Notifications
   * Specification.
   *
   * @param size the size to scale the icon
   *
   * @return a pixbuf or null if an icon can't be found
   */
  public Gdk.Pixbuf? load_icon(int size)
  {
    Gdk.Pixbuf pixbuf;

    /* If we have image data, use that */
    if(this._image_data != null) {
      pixbuf = this._image_data.to_pixbuf(size);
      
      if(pixbuf != null)
        return pixbuf;
    }

    /* Try to load the image-path */
    pixbuf = Icon.load(this._image_path, size);

    if(pixbuf != null)
      return pixbuf;

    /* Try to load the app_icon */
    pixbuf = Icon.load(this._app_icon, size);

    if(pixbuf != null)
      return pixbuf;

    /* Try to load the icon_data */
    if(this._icon_data != null) {
      pixbuf = this._icon_data.to_pixbuf(size);

      if(pixbuf != null)
        return pixbuf;
    }

    /* FIXME: Load a default icon here? */
    return null;
  }

  /**
   * The application name that sent this notification.
   */
  public string app_name {
    get { return this._app_name; }
  }

  /**
   * The name of an application icon for this notification, which may be
   * overridden by an icon included in the notification hints.
   */
  public string app_icon {
    get { return this._app_icon; }
  }

  /**
   * A brief summary of the notification.
   */
  public string summary {
    get { return this._summary; }
  }

  /**
   * The main body of text included in the notification.
   */
  public string body {
    get { return this._body; }
  }

  /**
   * The timestamp the Notification object was created. The Desktop
   * Notification specification unfortunately does not include a timestamp for
   * when the message was sent, but we can at least keep track of when we
   * received it.
   */
  public DateTime timestamp {
    get { return this._timestamp; }
  }

  /**
   * A notification is considered empty if after stripping the whitespace, both
   * the notification summary and body are empty. This is useful to discard
   * notifications which do not contain any useful information, such as volume
   * notifications.
   *
   * @return true if this notification is empty, false otherwise
   */
  public bool is_empty()
  {
    string summary = this._summary.strip();
    string body = this._body.strip();
    return (summary.length == 0) && (body.length == 0);
  }
}
