/**
 * Wrap GLib.Settings to provide preferences for the applet.
 */
public class Preferences : Object
{
  public const string SCHEMA_ID = "net.launchpad.RecentNotifications";
  public const string SCHEMA_PATH = "/net/launchpad/RecentNotifications/";

  /* Keys */
  public const string BLACKLIST          = "blacklist";
  public const string CONSOLE_LOG        = "console-log";
  public const string DEBUG              = "debug";
  public const string NOTIFICATION_LIMIT = "notification-limit";
  public const string SHOW_COUNT         = "show-count";

  private Settings _settings;

  public Preferences()
  {
    this._settings = new Settings(SCHEMA_ID);
  }

  /**
   * Binds an object's property. When the property is set, the GSettings key is
   * updated, and when the GSettings key is updated, the property is updated.
   *
   * @param key the preference key
   * @param object the object interested
   * @param property the object's property to bind
   */
  public void bind(string key, Object object, string property)
  {
    this._settings.bind(key, object, property, SettingsBindFlags.DEFAULT);
  }

  /**
   * Binds an object's property. When the GSettings key is updated, the
   * property is updated.
   *
   * @param key the preference key
   * @param object the object interested
   * @param property the object's property to bind
   */
  public void bind_read_only(string key, Object object, string property)
  {
    this._settings.bind(key, object, property, SettingsBindFlags.GET);
  }

  /**
   * Shows the preferences dialog.
   */
  public void show_dialog()
  {
    Gtk.Builder builder = new Gtk.Builder();
    builder.set_translation_domain(Config.GETTEXT_PACKAGE);
    string filename = Config.UI_DATA_DIR + "/preferences-dialog.ui";

    try {
      builder.add_from_file(filename);
    }
    catch(Error e) {
      warning("Failed to load preferences dialog from '%s': %s", filename, e.message);
      return;
    }

    Gtk.Dialog dialog = builder.get_object("preferences_dialog") as Gtk.Dialog;

    /* Bind the general settings */
    Gtk.CheckButton debug_button = builder.get_object("debug_button") as Gtk.CheckButton;
    this.bind(DEBUG, debug_button, "active");

    Gtk.CheckButton console_log_button = builder.get_object("console_log_button") as Gtk.CheckButton;
    this.bind(CONSOLE_LOG, console_log_button, "active");

    Gtk.CheckButton show_count_button = builder.get_object("show_count_button") as Gtk.CheckButton;
    this.bind(SHOW_COUNT, show_count_button, "active");

    Gtk.SpinButton notification_limit_button = builder.get_object("notification_limit_button") as Gtk.SpinButton;
    this.bind(NOTIFICATION_LIMIT, notification_limit_button, "value");

    /* Bind the blacklist settings */
    BlacklistPreference blacklist_preferences = new BlacklistPreference(this._settings, builder);
    blacklist_preferences.connect_handlers();

    dialog.run();
    dialog.destroy();
  }
}

/**
 * Maintain synchronization between the Gtk.ListStore and the GLib.Settings key
 * for the blacklist.
 */
private class BlacklistPreference : Object
{
  private Gtk.TreeView  _treeview;
  private Gtk.ListStore _liststore;

  private Gtk.Button _add_button;
  private Gtk.Button _remove_button;

  private Gtk.Entry _add_entry;

  private Settings _settings;

  private enum Column {
    APP_NAME = 0,
    COUNT
  }

  /**
   * Creates a new blacklist preference object and binds the blacklist setting.
   *
   * @param settings the GLib.Settings object to bind
   */
  public BlacklistPreference(Settings settings, Gtk.Builder builder)
  {
    this._liststore = new Gtk.ListStore(Column.COUNT,
      typeof(string) // APP_NAME
    );

    /* Have to create the TreeViewColumn manually...for some reason. */
    Gtk.TreeViewColumn column = new Gtk.TreeViewColumn();
    column.title = _("Application Name");

    this._treeview = builder.get_object("blacklist_treeview") as Gtk.TreeView;
    this._treeview.set_model(this._liststore);
    this._treeview.append_column(column);

    Gtk.CellRendererText cell = new Gtk.CellRendererText();
    column.pack_start(cell, true);
    column.add_attribute(cell, "text", Column.APP_NAME);

    this._add_entry = builder.get_object("blacklist_add_entry") as Gtk.Entry;
    this._add_button = builder.get_object("blacklist_add_button") as Gtk.Button;
    this._remove_button = builder.get_object("blacklist_remove_button") as Gtk.Button;

    this._settings = settings;
    this.set_values(settings.get_strv(Preferences.BLACKLIST));
  }

  /**
   * Connects the signal handlers and binds the properties. Using a separate
   * method so vala doesn't complain BlacklistPreferences is unused.
   */
  public void connect_handlers()
  {
    /* Add an item to the blacklist */
    this._add_button.clicked.connect((button) => {
      button.sensitive = false;

      string text = this._add_entry.get_text();
      this._add_entry.set_text("");
      text = text.strip();

      if(text.length > 0) {
        this.append_item(text);
        this._settings.set_strv(Preferences.BLACKLIST, this.get_values());
      }
      
      button.sensitive = true;
    });

    /* Remove an item from the blacklist */
    this._remove_button.clicked.connect((button) => {
      button.sensitive = false;

      if(this.remove_item()) {
        this._settings.set_strv(Preferences.BLACKLIST, this.get_values());
      }

      button.sensitive = true;
    });
  }

  /**
   * Appends an item to the blacklist.
   *
   * @param item the string to append
   */
  private void append_item(string item)
  {
    lock(this._liststore) {
      int i = this._liststore.iter_n_children(null);
      this._liststore.insert_with_values(null, i, Column.APP_NAME, item);
    }
  }

  /**
   * Removes the selected item from the blacklist.
   *
   * @return true if an item was removed
   */
  private bool remove_item()
  {
    lock(this._liststore) {
      Gtk.TreeSelection selection = this._treeview.get_selection();
      Gtk.TreeIter iter;
      
      if(selection.get_selected(null, out iter)) {
        this._liststore.remove(iter);
        return true;
      }
      else {
        return false;
      }
    }
  }

  /**
   * Updates a Gtk.ListStore of a single string column with the values from an
   * array of strings.
   *
   * @param values the values to insert
   */
  private void set_values(string[] values)
  {
    lock(this._liststore) {
      this._liststore.clear();

      for(int i = 0; i < values.length; i++)
        this._liststore.insert_with_values(null, i, Column.APP_NAME, values[i]);
    }
  }

  /**
   * Converts the values of a Gtk.ListStore of a single string column to an
   * array of strings.
   * 
   * @return a new array of strings
   */
  private string[] get_values()
  {
    lock(this._liststore) {
      int rows = this._liststore.iter_n_children(null);
      string[] values = new string[rows];
      Gtk.TreeIter iter;

      if(this._liststore.get_iter_first(out iter)) {
        int i = 0;
        do {
          string column_value;
          this._liststore.get(iter, Column.APP_NAME, out column_value);
          values[i++] = column_value;
        } while(this._liststore.iter_next(ref iter));
      }

      return values;
    }
  }
}
