/**
 * View recent notifications sent to a notification daemon. This is the main
 * applet code.
 */
public class RecentNotifications : Gtk.ToggleButton
{
  private Gtk.Image _image;
  private Gtk.Label _label;
  private Gtk.Box   _box;

  private Panel.Applet _applet;

  private NotificationMonitor _monitor;
  private NotificationWindow  _window;

  private Logging _logger;

  private Preferences _preferences;

  private bool _has_unread   = false;
  private int  _unread_count = 0;

  private const string ICON_READ   = "notification-read";
  private const string ICON_UNREAD = "notification-unread";

  private const string LOG_FILE_NAME = "recent-notifications.log";

  private const Gtk.IconSize ICON_SIZE = Gtk.IconSize.BUTTON;

  /**
   * Constructs a new Recent Notifications applet.
   *
   * @param applet the parent applet
   */
  public RecentNotifications(Panel.Applet applet)
  {
    /* Setup logging */
    this._logger = new Logging(LOG_FILE_NAME);

    this._preferences = new Preferences();
    this._preferences.bind_read_only(Preferences.DEBUG, this._logger, "log_debug");
    this._preferences.bind_read_only(Preferences.CONSOLE_LOG, this._logger, "log_to_console");

    debug("Creating applet");

    this._applet = applet;

    this.set_relief(Gtk.ReliefStyle.NONE);

    /* Main applet widgets */
    this._box = new Gtk.Box(Gtk.Orientation.HORIZONTAL, 4);
    this._box.show();
    this.add(this._box);

    this._image = new Gtk.Image.from_icon_name(ICON_READ, ICON_SIZE);
    this._image.show();
    this._box.pack_start(this._image);

    this._label = new Gtk.Label(_("0"));
    this._preferences.bind_read_only(Preferences.SHOW_COUNT, this._label, "visible");
    this._box.pack_start(this._label);

    /* Don't eat the button press */
    this.button_press_event.connect((widget, event) => {
      if(event.button != 1)
        Signal.stop_emission_by_name(widget, "button_press_event");

      return false;
    });

    this._window = new NotificationWindow();
    this._preferences.bind_read_only(Preferences.NOTIFICATION_LIMIT, this._window,
        "notification_limit");

    /* Watch for notifications */
    this._monitor = new NotificationMonitor();
    this._preferences.bind_read_only(Preferences.BLACKLIST, this._monitor,
        "blacklist");
    this._monitor.received.connect((message) => {
      this._window.add_notification(message);
      this.set_unread(true);
    });

    /* Show/Hide the dock window when the toggle button is clicked */
    this.toggled.connect((button) => {
      if(button.active)
        this.show_window();
      else
        this.hide_window();
    });

    /* Update applet on orientation changes */
    this.update_orient(applet.get_orient());
    applet.change_orient.connect((orient) => this.update_orient(orient));

    this.create_menu();

    this.show();
  }

  /**
   * Creates the applet context menu.
   */
  private void create_menu()
  {
    Gtk.ActionGroup group = new Gtk.ActionGroup("applet");
    Gtk.Action action;

    action = new Gtk.Action("Clear", null, null, Gtk.Stock.CLEAR);
    action.activate.connect(this.clear_cb);
    group.add_action(action);

    action = new Gtk.Action("About", null, null, Gtk.Stock.ABOUT);
    action.activate.connect(this.about_cb);
    group.add_action(action);

    action = new Gtk.Action("Preferences", null, null, Gtk.Stock.PREFERENCES);
    action.activate.connect(this.preferences_cb);
    group.add_action(action);

    debug("UI Directory: %s", Config.UI_DATA_DIR);
    this._applet.setup_menu_from_file(Config.UI_DATA_DIR + "/applet-menu.xml", group);
  }

  /* CALLBACKS */

  /**
   * Shows the about dialog.
   *
   * @param action the action that was activated [not used]
   */
  private void about_cb(Gtk.Action action)
  {
    About dialog = new About();
    dialog.run();
    dialog.destroy();
  }

  /**
   * Clears the notification list.
   *
   * @param action the action that was activated [not used]
   */
  private void clear_cb(Gtk.Action action)
  {
    this._window.clear_notifications();
    this.set_unread(false);
  }

  /**
   * Shows the preferences dialog.
   *
   * @param action the action that was activated [not used]
   */
  private void preferences_cb(Gtk.Action action)
  {
    this._preferences.show_dialog();
  }

  /**
   * Shows the applet's dock window.
   */
  private void show_window()
  {
    Gdk.Window window = this.get_window();

    int x, y;
    window.get_origin(out x, out y);

    int width = window.get_width();
    int height = window.get_height();

    Gdk.Screen screen = this.get_screen();

    Panel.AppletOrient orient = this._applet.get_orient();

    this._window.align_with_parent(x, y, width, height, screen, orient);
    this._window.show();
  }

  /**
   * Hides the applet's dock window, and clears the unread notification count.
   */
  private void hide_window()
  {
    this._window.hide();
    this.set_unread(false);
  }

  /**
   * Resets the orientation of the applet based on the orientation of the panel
   * which contains it.
   *
   * @param orient the new orientation
   */
  private void update_orient(Panel.AppletOrient orient)
  {
    if(orient == Panel.AppletOrient.UP || orient == Panel.AppletOrient.DOWN) {
      this._box.set_orientation(Gtk.Orientation.HORIZONTAL);
      this._label.set_angle(0);
    }
    else {
      this._box.set_orientation(Gtk.Orientation.VERTICAL);
      this._label.set_angle(90);
    }
  }

  /**
   * Sets the unread status of the notifications and updates the notification
   * count and icon depending on the status.
   *
   * @param unread the status of unread notifications
   */
  private void set_unread(bool unread)
  {
    this._has_unread = unread;

    string icon_name;
    Gtk.IconSize size;
    this._image.get_icon_name(out icon_name, out size);

    if(unread) {
      /* Increment the notification count */
      this._unread_count++;
      if(icon_name == ICON_READ)
        this._image.set_from_icon_name(ICON_UNREAD, size);
    }
    else {
      /* Reset the notification count */
      this._unread_count = 0;
      if(icon_name == ICON_UNREAD)
        this._image.set_from_icon_name(ICON_READ, size);
    }

    this._label.set_text(ngettext("1", "%d", this._unread_count).printf(this._unread_count));
  }
}
